#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <jcPKCS11.h>

// используем имеющиеся классы для совместимости с SDK
#include "P11Loader.h"
#include "Utils.h"

#define CSR_FILE            "csr.der"
#define PIN_DEFAULT_USER    "1234567890"
#define MAX_SEARCH_COUNT    100

static P11Loader *p11;

#define VARATTR_METHOD(ATTR, TYPE) \
static TYPE * \
get##ATTR(CK_SESSION_HANDLE sess, CK_OBJECT_HANDLE obj, CK_ULONG_PTR pulCount) \
{ \
    P11Loader &loader = *p11;                                               \
	CK_ATTRIBUTE attr = { CKA_##ATTR, NULL, 0 };                            \
	CK_RV rv;                                                               \
	if (pulCount)                                                           \
		*pulCount = 0;                                                      \
	rv = CALL_P11(C_GetAttributeValue(sess, obj, &attr, 1));                \
	if (rv == CKR_OK) {                                                     \
		if (attr.ulValueLen == (CK_ULONG)(-1))                              \
			return NULL;                                                    \
		if (!(attr.pValue = calloc(1, attr.ulValueLen + 1))) {              \
			fprintf(stderr, "out of memory in get" #ATTR ": %m");           \
            exit(EXIT_FAILURE);                                             \
        }                                                                   \
		rv = CALL_P11(C_GetAttributeValue(sess, obj, &attr, 1));            \
		if (attr.ulValueLen == (CK_ULONG)(-1)) {                            \
			free(attr.pValue);                                              \
			return NULL;                                                    \
		}                                                                   \
		if (pulCount)                                                       \
			*pulCount = attr.ulValueLen / sizeof(TYPE);                     \
	} else {                                                                \
		fprintf(stderr, "C_GetAttributeValue(" #ATTR ")");                  \
	}                                                                       \
	return (TYPE *) attr.pValue;                                            \
}

VARATTR_METHOD(LABEL, char);				/* getLABEL */
VARATTR_METHOD(ID, unsigned char);			/* getID */

static CK_SESSION_HANDLE    hsession = 0;

static void display_usage(void);
static int find_keys(CK_OBJECT_HANDLE_PTR pkeys);
static void show_key(CK_OBJECT_HANDLE key);
static int parse_DN(char *str, CK_CHAR_PTR_PTR *dn);
static int parse_exts(char *str, CK_CHAR_PTR_PTR* exts);
static int replace_char(char* str, char oldc, char newc);

struct global_args_t {
    char *module;
    char *dn;
    char *exts;
    char *csr_file;
    char *pin;
} global_args;

static const char *opt_str = "M:p:n:x:o:h";

static const struct option options[] = {
    { "module", required_argument, NULL, 'M' },
    { "pin", required_argument, NULL, 'p' },
    { "output-file", required_argument, NULL, 'o'},
    { "dn", required_argument, NULL, 'n'},
    { "exts", required_argument, NULL, 'x' },
    { NULL, no_argument, NULL, 0 }
};

int main(int argc, char *argv[]) {
    int                 opt = 0;
    int                 optind = 0;
    int                 i = 0;

    CK_RV               rv = CKR_OK;
    CK_SLOT_ID_PTR      pslots = NULL;      // массив слотов
    CK_ULONG            nslots = 0;         // кол-во слотов
    CK_TOKEN_INFO       token_info = {};
    bool                gost2_found = false;
    CK_SLOT_ID          gost2_slot_id = 0;
    CK_OBJECT_HANDLE    hpubkey = 0;
    CK_OBJECT_HANDLE    hprivkey = 0;       // поиск закрытого ключа по открытому
    CK_OBJECT_HANDLE    pkeys[MAX_SEARCH_COUNT];   // массив ключей
    int                 nkeys = 0;          // кол-во найденных ключей
    FILE*               fpcsr = NULL;       // файл с запросом на сертификат
    CK_BYTE_PTR         pcsr = NULL;        // запрос на сертификат
    CK_ULONG            csrlen = 0;         // длина запроса на сертификат
    CK_CHAR_PTR         pin = (CK_CHAR_PTR) PIN_DEFAULT_USER;
    CK_CHAR_PTR_PTR     dn = NULL;
    int                 dnsize = 0;
    CK_CHAR_PTR_PTR     exts = NULL;
    int                 exts_size = 0;

    global_args.module = NULL;
    global_args.dn = NULL;
    global_args.exts = NULL;
    global_args.csr_file = CSR_FILE;

    while ((opt = getopt_long(argc, argv, opt_str, options, &optind)) != -1) {
        switch (opt) {
            case 'M':
                global_args.module = optarg;
                break;
            case 'n':
                global_args.dn = optarg;
                break;
            case 'x':
                global_args.exts = optarg;
                break;
            case 'p':
                global_args.pin = optarg;
                break;
            case 'o':
                global_args.csr_file = optarg;
                break;
            case 'h':
                display_usage();
                return EXIT_SUCCESS;
            case '?':
            default:
                return EXIT_FAILURE;
        }
    }

    if (!global_args.module) {
        fprintf(stderr, "PKCS #11 module is required, use: --module option\n");
        return -1;
    }

    if (!global_args.dn) {
        fprintf(stderr, "DN string is required\n");
        return -1;
    }

    if ((dnsize = parse_DN(global_args.dn, &dn)) == -1) {
        return -1;
    }

    if (global_args.exts != NULL)
        if ((exts_size = parse_exts(global_args.exts, &exts)) == -1) {
            return -1;
        }
    
    if (global_args.pin != NULL)
        pin = (CK_CHAR_PTR)global_args.pin;

    P11Loader loader(global_args.module);
    if (!loader.IsInitialized())
        return EXIT_FAILURE;

    p11 = &loader;

    rv = CALL_P11(C_Initialize(0));
    if (rv != CKR_OK) {
        P11Utils::PrintError("C_Initialize", rv);
        goto end;
    }

    printf("jcPKCS11-2 library initialized.\n");

    // получаем список слотов; в стандарте pkcs #11 под cписком слотов понимается список ридеров.
    // параметр CK_TRUE указывает на то, что мы запрашиваем ридеры, в которые вставлены смарт-карты.
    // в случае USB токенов в одном устройстве комбинируется и ридер и смарт-карта.
    // в случае многоапплетных токенов одному токену может соответствовать несколько слотов pkcs #11.
    rv = CALL_P11(C_GetSlotList(CK_TRUE, NULL, &nslots));
    if (rv != CKR_OK) {
        P11Utils::PrintError("C_GetSlotList", rv);
        goto end;
    }

    // таких слотов нет?
    if (nslots == 0) {
        printf("No tokens are present.\n");
        goto end;
    }

    pslots = (CK_SLOT_ID_PTR)malloc(sizeof(CK_SLOT_ID) * nslots);

    rv = CALL_P11(C_GetSlotList(CK_TRUE, pslots, &nslots));
    if (rv != CKR_OK) {
        P11Utils::PrintError("C_GetSlotList", rv);
        goto end;
    }

    // поиск апплета ГОСТ2
    for (CK_ULONG i = 0; i < nslots; ++i) {
        // получение информации о токене в слоте
        // в стандарте pkcs #11 под токеном понимается апплет на смарткарте, вставленной в ридер
        rv = CALL_P11(C_GetTokenInfo(pslots[i], &token_info));
        if (rv != CKR_OK)
        {
            P11Utils::PrintError("C_GetSlotInfo", rv);
            goto end;
        }

        // проверка наличия апплета ГОСТ2
        bool bFound = !memcmp((const char*) token_info.model, JC_MODEL_CRYPTOTOKEN_2, strlen(JC_MODEL_CRYPTOTOKEN_2));
        if (bFound)
        {
            gost2_found = true;
            gost2_slot_id = pslots[i];

            printf("Applet GOST2 found in slot %lu\n", gost2_slot_id);
            break;
        }
    }

    if (!gost2_found) {
        fprintf(stderr, "No smartcard with GOST2 applet found.\n");
        goto end;
    }

    // открываем сессию
    rv = CALL_P11(C_OpenSession(gost2_slot_id, (CKF_SERIAL_SESSION | CKF_RW_SESSION), 0, 0, &hsession));
    if (rv != CKR_OK) {
        P11Utils::PrintError("C_OpenSession", rv);
        goto end;
    }

    printf("Open session: OK\n");

    // подключаемся к токену с PIN-кодом по умолчанию
    rv = CALL_P11(C_Login(hsession, CKU_USER, pin, strlen((const char*)pin)));
    if (rv != CKR_OK) {
        P11Utils::PrintError("C_Login", rv);
        goto end;
    }
    printf("Login: OK\n");

    // Ищем публичный ключ
    nkeys = find_keys(pkeys);
    if (nkeys < 1) {
        fprintf(stderr, "No public keys found. Please, generate a keypair");
        goto end;
    }

    for (i=0; i < nkeys; i++) {
        printf("Key %d:\n", i+1);
        show_key(pkeys[i]);
    }

    if (nkeys == 1) {
        hpubkey = pkeys[0];
    } else {
        printf("Please, select public key: ");
        if (!scanf("%d", &i) || i > nkeys || i < 1) {
            fprintf(stderr, "Bad key number\n");
            goto end;
        }

        hpubkey = pkeys[i-1];
    }

    rv = CALL_EXT(createCSR(
        hsession,
        hpubkey,
        dn, dnsize,
        &pcsr, &csrlen,
        hprivkey, NULL, 0,
        exts, exts_size));
    if (rv != CKR_OK) {
        P11Utils::PrintError("createCSR", rv);
        goto end;
    }
    printf("Create CSR: OK\n");

    // создаем файл для записи запроса на сертификат
    fpcsr = fopen(global_args.csr_file, "wb");
    if (fpcsr == NULL) {
        printf("Failed to create csr file.");
        goto end;
    }

    // записываем запрос на сертификат в файл
    fwrite(pcsr, 1, csrlen, fpcsr);

    // закрываем файл
    fclose(fpcsr);

    printf("Write CSR to file: OK\n");
    
end:
    if (hsession != 0) {
        // завершаем сеанс пользователя, если он был авторизован
        CALL_P11(C_Logout(hsession));
        //закрываем сессию
        CALL_P11(C_CloseSession(hsession));
    }

    // освобождаем память
    if (pcsr)
        CALL_EXT(freeBuffer(pcsr));

    if (pslots != NULL)
        free(pslots);

    if (dn != NULL)
        free(dn);
    
    if (exts != NULL)
        free(exts);
    
    // заканчиваем работу с библиотекой
    CALL_P11(C_Finalize(0));
    return EXIT_SUCCESS;
}

void display_usage(void) {
    puts("jc_certreq - generate certificate request");
}

int find_keys(CK_OBJECT_HANDLE_PTR pkeys) {
    P11Loader &loader = *p11;

    CK_RV rv;
    CK_ULONG key_class = CKO_PUBLIC_KEY;
    CK_BBOOL btrue = CK_TRUE;
    // CK_BBOOL bfalse = CK_TRUE;
    CK_ULONG nfound;

    CK_ATTRIBUTE search_attrs[] =
    {
        { CKA_CLASS, &key_class, sizeof(key_class) },
        { CKA_TOKEN, &btrue, sizeof(btrue) },
    };

    rv = CALL_P11(C_FindObjectsInit(
        hsession, search_attrs, sizeof(search_attrs) / sizeof(CK_ATTRIBUTE)));
    if (rv != CKR_OK) {
        P11Utils::PrintError("C_FindObjectsInit", rv);
        return -1;
    }

    rv = CALL_P11(C_FindObjects(hsession, pkeys, MAX_SEARCH_COUNT, &nfound));
    if (rv != CKR_OK) {
        P11Utils::PrintError("C_FindObjects", rv);
        return -1;
    }

    rv = CALL_P11(C_FindObjectsFinal(hsession));
    if (rv != CKR_OK) {
        P11Utils::PrintError("C_FindObjectsFinal", rv);
        return -1;
    }

    return nfound;
}

void show_key(CK_OBJECT_HANDLE key) {
    CK_ULONG size = 0;
    unsigned char *id;
    char *label;

    if ((label = getLABEL(hsession, key, NULL)) != NULL) {
		printf("\tLabel: %s\n", label);
		free(label);
	}

    if ((id = getID(hsession, key, &size)) != NULL && size) {
        unsigned int n;

        printf("\tID: ");
        for (n = 0; n < size; n++)
            printf("%02x", id[n]);
        printf("\n");
        free(id);
    }
}

int parse_DN(char *str, CK_CHAR_PTR_PTR *dn) {
    char *pc, *ps, *peq, *pkey, *pval;
    size_t len = 0, dnsize=0;
    int npairs = 1;

    if ((len = strlen(str)) < 3 || str[0] == ',') {
        fprintf(stderr, "Invalid DN\n");
        return -1;
    }

    // определим кол-во пар и заменим запятые на '\0'
    for (ps = (char*)str; (pc = strchr(ps, ',')) != NULL; ps = pc+1) {
        if (*(pc-1) == '\\') {
            // снимаем экранирование с запятых
            *(pc-1) = ',';
            *pc = ' ';
        } else {
            *pc = '\0';
            ++npairs;
        }
    }

    if (!(*dn = (CK_CHAR_PTR*) malloc(2 * npairs * sizeof(CK_CHAR_PTR)))) {
        fprintf(stderr, "Out of memory\n");
        return -1;
    }

    for (ps = str; npairs > 0; --npairs, ps += len + 1) {
        if (!(peq = strchr(ps, '='))) {
            fprintf(stderr, "Invalid DN: '=' expected near with '%s'\n", ps);
            free(*dn);
            return -1;
        }

        len = strlen(ps);
        *peq = '\0';

        if (peq == ps || peq == ps + len - 1) {
            fprintf(stderr, "Invalid DN: key or value must not be empty near with '%s'\n", ps);
            free(*dn);
            return -1;
        }

        pkey = ps;
        pval = peq + 1;

        (*dn)[dnsize++] = (CK_CHAR_PTR)pkey;
        (*dn)[dnsize++] = (CK_CHAR_PTR)pval;
    }

    return dnsize;
}

int parse_exts(char *str, CK_CHAR_PTR_PTR* exts) {
    int npairs = 0;
    size_t pslen = 0, exts_size = 0;
    char *ps, *peq, *pkey, *pval;
    npairs = replace_char(str, ';', '\0') + 1;

    if (!(*exts = (CK_CHAR_PTR*) malloc(2 * npairs * sizeof(CK_CHAR_PTR)))) {
        fprintf(stderr, "Out of memory\n");
        return -1;
    }

    for (ps = str; npairs > 0; --npairs, ps += pslen + 1) {
        if (!(peq = strchr(ps, '='))) {
            fprintf(stderr, "Invalid extension: '=' expected near with '%s'\n", ps);
            free(*exts);
            return -1;
        }

        pslen = strlen(ps);
        *peq = '\0';

        if (peq == ps || peq == ps + pslen - 1) {
            fprintf(stderr, "Invalid extension: key or value must not be empty near with '%s'\n", ps);
            free(*exts);
            return -1;
        }

        pkey = ps;
        pval = peq + 1;

        (*exts)[exts_size++] = (CK_CHAR_PTR)pkey;
        (*exts)[exts_size++] = (CK_CHAR_PTR)pval;
    }

    return exts_size;
}

int replace_char(char* str, char oldc, char newc) {
    int n = 0;
    char *ps, *pc;
    for (ps = (char*)str; (pc = strchr(ps, oldc)) != NULL; ps = pc + 1, ++n)
        *pc = newc;
    return n;
}